/*
 
File:	gfx2d.h

Purpose: API for 2D Graphics Engine

*/

// We are using display controller set to 640x480
// Modify as necessary if using different resolution
// --> Also be sure to change `defines in gfx2d.v and in display controller (SVGA_TIMING_GENERATION.v)
#define SCREEN_W 640
#define SCREEN_H 480

// Some colour defines
#define BLACK 	0x000000
#define WHITE 	0xFFFFFF
#define RED 	0xFF0000
#define GREEN 	0x00FF00
#define BLUE 	0x0000FF
#define YELLOW	0xFFFF00
#define VIOLET	0xFF00FF
#define CYAN 	0x00FFFF

// Masks for gfx2d status register
#define OP_BIT			(0x7)
#define REQ_BIT			(0x1 << 3)
#define BLIT_RTR_BIT	(0x1 << 4)
#define LINE_RTR_BIT	(0x1 << 5)
#define CHAR_RTR_BIT	(0x1 << 6)
#define GFX_RTR_BIT		(0x1 << 7)

// A smaller footprint printf
#define printf xil_printf

// Function prototypes
int lineReady();
int blitReady();
int gfx2dReady();
int setData (Xuint32 x0, Xuint32 y0, Xuint32 x1, Xuint32 y1, Xuint32 RGB);
int lineOp (Xuint32 x0, Xuint32 y0, Xuint32 x1, Xuint32 y1, Xuint32 RGB);
int blitOp (Xuint32 x0, Xuint32 y0, Xuint32 x1, Xuint32 y1, Xuint32 RGB);
int charOp (Xuint32 x0, Xuint32 y0, Xuint32 ascii, Xuint32 RGB);
int pixelOp (Xuint32 x, Xuint32 y, Xuint32 RGB);
int gfxRegRead ();

int clearScreen (Xuint32 RGB);
int drawRect (Xuint32 x0, Xuint32 y0, Xuint32 width, Xuint32 height, Xuint32 RGB);
int fillRect (Xuint32 x0, Xuint32 y0, Xuint32 width, Xuint32 height, Xuint32 RGB);
int drawSquare (Xuint32 x0, Xuint32 y0, Xuint32 length, Xuint32 RGB);
int fillSquare (Xuint32 x0, Xuint32 y0, Xuint32 length, Xuint32 RGB);
int drawTriangle (Xuint32 x0, Xuint32 y0, Xuint32 x1, Xuint32 y1, Xuint32 x2, Xuint32 y2, Xuint32 RGB);
int drawStar (Xuint32 x, Xuint32 y, Xuint32 r1, Xuint32 r2, Xuint32 RGB);
int rotateSquare90 (Xuint32 x, Xuint32 y, Xuint32 r, Xuint32 RGB, int clear);
int rotateStar90 (Xuint32 x, Xuint32 y, Xuint32 r1, Xuint32 r2, Xuint32 RGB, int clear);
void drawString (Xuint32 x, Xuint32 y, Xuint32 RGB, char* msg);

void ppmOp (Xuint32 x, Xuint32 y, Xuint8 * ppm);
void stringToInt (Xuint8 * string, Xuint32 * decimal);
void delay(int max);

// ---------------------------
// Low level Gfx 2D operations
// ---------------------------

// Masks status register to determine if graphics engine ready 
int gfx2dReady() {
	Xuint32 status;

	status = XIo_In32(XPAR_GFX2D_0_BASEADDR);
	return ((status & GFX_RTR_BIT) ? 1 : 0);
}


// Set input data for gfx2d operations
int setData (Xuint32 x0, Xuint32 y0, Xuint32 x1, Xuint32 y1, Xuint32 RGB) {

	// Register writes to set colour and coordinate data
	XIo_Out32(XPAR_GFX2D_0_BASEADDR + 4,  RGB);
	XIo_Out32(XPAR_GFX2D_0_BASEADDR + 8,  x0);
	XIo_Out32(XPAR_GFX2D_0_BASEADDR + 12, y0);
	XIo_Out32(XPAR_GFX2D_0_BASEADDR + 16, x1);
	XIo_Out32(XPAR_GFX2D_0_BASEADDR + 20, y1);

	return 0;
}

// Line operation
int lineOp (Xuint32 x0, Xuint32 y0, Xuint32 x1, Xuint32 y1, Xuint32 RGB) {

	// Wait for gfx2d ready
	while (!gfx2dReady());

	// send line request
	setData(x0,y0,x1,y1,RGB);
	XIo_Out32(XPAR_GFX2D_0_BASEADDR, 8);

	return 0;
}

// Blit operation
int blitOp (Xuint32 x0, Xuint32 y0, Xuint32 x1, Xuint32 y1, Xuint32 RGB) {

	// Wait for gfx2d ready
	while (!gfx2dReady());

	// send blit request
	setData(x0,y0,x1,y1,RGB);
	XIo_Out32(XPAR_GFX2D_0_BASEADDR, 9);

	return 0;
}

// Char operation
int charOp (Xuint32 x0, Xuint32 y0, Xuint32 ascii, Xuint32 RGB) {

	// Wait for gfx2d ready
	while (!gfx2dReady());

	// send char request
	setData(x0,y0,ascii,0,RGB);
	XIo_Out32(XPAR_GFX2D_0_BASEADDR, 10);

	return 0;
}

// Pixel operation
int pixelOp (Xuint32 x, Xuint32 y, Xuint32 RGB) {

	// Wait for gfx2d ready
	while (!gfx2dReady());
	
	// send pixel request
	setData(x,y,0,0,RGB);
	XIo_Out32(XPAR_GFX2D_0_BASEADDR, 11);

	return 0;
}


// Register reads from gfx2d
int gfxRegRead () {

	Xuint32 data;
	int i;

	for (i = 0; i < 7; i++) {
		data = XIo_In32(XPAR_GFX2D_0_BASEADDR + 4*i);
		printf("\n\ni = %d, Read Data = %d\n\r", i, data);
	}

	return 0;
}


// ---------------------------
// High level operations
// ---------------------------

// Fills entire screen
int clearScreen (Xuint32 RGB) {
	return blitOp(0,0,SCREEN_W,SCREEN_H,RGB);
}

// Draw rectangle starting at coordinate x0,y0 with given width, height and colour RGB
int drawRect (Xuint32 x0, Xuint32 y0, Xuint32 width, Xuint32 height, Xuint32 RGB) {

	Xuint32 x1, y1;

	// Calculate endpoint
	x1 = x0 + width;
	y1 = y0 + height;

	// Draw line for each side of rectangle
	lineOp(x0, y0, x1, y0, RGB);
	lineOp(x0, y0, x0, y1, RGB);
	lineOp(x1, y0, x1, y1, RGB);
	lineOp(x0, y1, x1, y1, RGB);

	return 0;
}

// Fill rectangle starting at coordinate x0,y0 with given width, height and colour RGB
int fillRect (Xuint32 x0, Xuint32 y0, Xuint32 width, Xuint32 height, Xuint32 RGB) {
	return blitOp(x0, y0, x0+width, y0+height, RGB);
}

// Draw square starting at coordinate x0,y0 with given side length and colour RGB
int drawSquare (Xuint32 x0, Xuint32 y0, Xuint32 length, Xuint32 RGB) {
	return drawRect(x0, y0, length, length, RGB);
}

// Fill square starting at coordinate x0,y0 with given side length and colour RGB
int fillSquare (Xuint32 x0, Xuint32 y0, Xuint32 length, Xuint32 RGB) {
	return fillRect(x0, y0, length, length, RGB);
}

// Draw triangle given coordinates (x0,y0),(x1,y1),(x2,y2) and colour RGB
int drawTriangle (Xuint32 x0, Xuint32 y0, Xuint32 x1, Xuint32 y1, Xuint32 x2, Xuint32 y2, Xuint32 RGB) {

	// Draw line for each side of triangle
	lineOp(x0, y0, x1, y1, RGB);
	lineOp(x0, y0, x2, y2, RGB);
	lineOp(x1, y1, x2, y2, RGB);
	return 0;
}

// Draw four-pointed star given:
// - center (x,y)
// - r1: distance from center to "virtual square" bounding inside of star
// - r2: distance from center to endpoint of star
int drawStar (Xuint32 x, Xuint32 y, Xuint32 r1, Xuint32 r2, Xuint32 RGB) {
	
	// Draw lines for each side of star
	lineOp(x-r1, y-r1, x, y-r2, RGB);
	lineOp(x+r1, y-r1, x, y-r2, RGB);

	lineOp(x+r1, y-r1, x+r2, y, RGB);
	lineOp(x+r1, y+r1, x+r2, y, RGB);

	lineOp(x-r1, y+r1, x, y+r2, RGB);
	lineOp(x+r1, y+r1, x, y+r2, RGB);

	lineOp(x-r1, y-r1, x-r2, y, RGB);
	lineOp(x-r1, y+r1, x-r2, y, RGB);

	return 0;
}

// Draw square at coordinate x,y with given side length and colour RGB
// Rotate square 90 degrees
// clear: set to clear square from previous frame (set to 0 to get cool looking shapes)
int rotateSquare90 (Xuint32 x, Xuint32 y, Xuint32 r, Xuint32 RGB, int clear) {
	/*
	r := r/sqrt(2)

	alpha (i*pi/20)	sin-cos		sin+cos		sin			cos
	0				-1			1			0
	1				-0.831254	1.144123	0.156434
	2				-0.642040	1.260074	0.309017
	3				-0.437016	1.344997	0.453990
	4				-0.221232	1.396802	0.587785
	5				0			1.414214	0.707107
	6				0.221232	1.396802	0.809017
	7				0.437016	1.344997	0.891007
	8				0.642040	1.260074	0.951057
	9				0.831254	1.144123	0.987688
	10				1			1			1			0

	i = 0:10
	x0(i) = x + r * sin-cos(i)
	y0(i) = y - r * sin+cos(i)

	x1(i) = x + r * sin+cos(i)
	y1(i) = y - r * cos-sin(i) = y + r * sin-cos(i)
	*/

	int i,j;

	// Hard code sin(a)-cos(a), sin(a)+cos(a), a = 0 to pi/2
	double sin_minus_cos [] = {-1, -0.831254, -0.642040, -0.437016, -0.221232, 0, 0.221232, 0.437016, 0.642040, 0.831254, 1};
	double sin_plus_cos [] = {1, 1.144123, 1.260074, 1.344997, 1.396902, 1.414214, 1.396802, 1.344997, 1.260074, 1.144123, 1};

	Xuint32 x0, y0, x1, y1, x2, y2, x3, y3;

	// Loop to rotate 90 degrees
	for (i = 0; i <= 10; i++) {
		double r_plus, r_minus;
		r_plus  = (double)r * sin_plus_cos[i];
		r_minus = (double)r * sin_minus_cos[i];

		// Clear previous square
		if (clear) {
			//lineOp(x0, y0, x1, y1, BLACK);
			//lineOp(x1, y1, x2, y2, BLACK);
			//lineOp(x2, y2, x3, y3, BLACK);
			//lineOp(x3, y3, x0, y0, BLACK);
			blitOp(x-1.5*r-1,y-1.5*r-1,x+1.5*r+1,y+1.5*r+1,BLACK);
		}

		// Calculate next endpoint locations
		x0 = x + (Xuint32)r_minus;
		y0 = y - (Xuint32)r_plus;
		x1 = x + (Xuint32)r_plus;
		y1 = y + (Xuint32)r_minus;
		x2 = x - (Xuint32)r_minus;
		y2 = y + (Xuint32)r_plus;
		x3 = x - (Xuint32)r_plus;
		y3 = y - (Xuint32)r_minus;

		// Draw lines for new square
		lineOp(x0, y0, x1, y1, RGB);
		lineOp(x1, y1, x2, y2, RGB);
		lineOp(x2, y2, x3, y3, RGB);
		lineOp(x3, y3, x0, y0, RGB);

		//delay(2);
	}

}

// Draw four-pointed star given:
// - center (x,y)
// - r1: distance from center to "virtual square" bounding inside of star
// - r2: distance from center to endpoint of star
// Rotate star 90 degrees
// clear: set to clear star from previous frame (set to 0 to get cool looking shapes)
int rotateStar90 (Xuint32 x, Xuint32 y, Xuint32 r1, Xuint32 r2, Xuint32 RGB, int clear) {

	int i,j;

	// Hard code sin(a), cos(a), sin(a)-cos(a), sin(a)+cos(a), a = 0 to pi/2
	double sin [] = {0, 0.156434, 0.309017, 0.453990, 0.587785, 0.707107, 0.809017, 0.891007, 0.951057, 0.987688, 1};
	double cos [] = {1, 0.987688, 0.951057, 0.891007, 0.809017, 0.707107, 0.587785, 0.453990, 0.309017, 0.156434, 0};
	double sin_minus_cos [] = {-1, -0.831254, -0.642040, -0.437016, -0.221232, 0, 0.221232, 0.437016, 0.642040, 0.831254, 1};
	double sin_plus_cos [] = {1, 1.144123, 1.260074, 1.344997, 1.396902, 1.414214, 1.396802, 1.344997, 1.260074, 1.144123, 1};

	Xuint32 x0, y0, x1, y1, x2, y2, x3, y3; // corners of inner square
	Xuint32 x4, y4, x5, y5, x6, y6, x7, y7; // end points of star

	// rotate 90 degrees
	for (i = 0; i <= 10; i++) {
		double r1_plus, r1_minus, r2_sin, r2_cos;
		r1_plus  = (double)r1 * sin_plus_cos[i];
		r1_minus = (double)r1 * sin_minus_cos[i];
		r2_sin   = (double)r2 * sin[i];
		r2_cos   = (double)r2 * cos[i];

		// Clear previous star
		if (clear) {
			blitOp(x-r2-1,y-r2-1,x+r2+1,y+r2+1,BLACK);
		}

		// Calculate next endpoint locations
		x0 = x + (Xuint32)r1_minus;
		y0 = y - (Xuint32)r1_plus;
		x1 = x + (Xuint32)r1_plus;
		y1 = y + (Xuint32)r1_minus;
		x2 = x - (Xuint32)r1_minus;
		y2 = y + (Xuint32)r1_plus;
		x3 = x - (Xuint32)r1_plus;
		y3 = y - (Xuint32)r1_minus;

		x4 = x + (Xuint32)r2_sin;
		y4 = y - (Xuint32)r2_cos;
		x5 = x + (Xuint32)r2_cos;
		y5 = y + (Xuint32)r2_sin;
		x6 = x - (Xuint32)r2_sin;
		y6 = y + (Xuint32)r2_cos;
		x7 = x - (Xuint32)r2_cos;
		y7 = y - (Xuint32)r2_sin;

		// Draw lines for new star
		lineOp(x0, y0, x4, y4, RGB);
		lineOp(x4, y4, x1, y1, RGB);

		lineOp(x1, y1, x5, y5, RGB);
		lineOp(x5, y5, x2, y2, RGB);

		lineOp(x2, y2, x6, y6, RGB);
		lineOp(x6, y6, x3, y3, RGB);
		
		lineOp(x3, y3, x7, y7, RGB);
		lineOp(x7, y7, x0, y0, RGB);

		delay(6);
	}
}

// Draws a string to screen using character operation
void drawString (Xuint32 x, Xuint32 y, Xuint32 RGB, char* msg) {
	int i = 0;
	while (msg[i]) {
		// Draw each character with spacing of 7 pixels
		charOp (x + i*7, y, msg[i], RGB);
		i++;
	}
}

// Macro to switch endianness of a byte
#define BIT_FLIP(x) (((x & 1) << 7) | \
					 ((x & 2) << 5) | \
					 ((x & 4) << 3) | \
					 ((x & 8) << 1) | \
					 ((x & 16) >> 1) | \
					 ((x & 32) >> 3) | \
					 ((x & 64) >> 5) | \
					 ((x & 128) >> 7))

// Reads in ppm data from an Xuint8/char array and displays to screen at (x,y)
void ppmOp (Xuint32 x, Xuint32 y, Xuint8 * ppm) {

	Xuint32 width, height;
	int i = 0, j = 0, k=0;
	Xuint8 buf[16], * ppm_info[4];
	Xuint8 * image = ppm, pixel[4];

	pixel[0] = 0;

	ppm_info[0] = buf;

	// Extract header data
	while (1) {
		buf[i] = *image++;

		if (buf[i] == ' ') {
			buf[i] = 0;
			j++;
			ppm_info[j] = &buf[i+1];
		}

		if (buf[i] == 0x0A)
			break;
			
		i++;
	}

	// Obtain width/height data
	stringToInt(ppm_info[1], &width);
	stringToInt(ppm_info[2], &height);

	// Go from 0:height-1 or height-1:0 depending on order of RGB data in ppm file
	//for (i = 0; i < height; i++) {
	for (i = height-1; i >= 0; i--) {
		for (j = 0; j < width; j++) {
			pixel[3] = BIT_FLIP(*image++);
			pixel[2] = BIT_FLIP(*image++);
			pixel[1] = BIT_FLIP(*image++);
			//pixelOp(x + j, y + i, *(Xuint32 *)pixel);
			blitOp(x + j, y + i, x + j + 1, y + i + 1, *(Xuint32 *)pixel);
		}		
	}

}

// Helper function for ppmOp
void stringToInt (Xuint8 * string, Xuint32 * decimal) {
	Xuint32 value = 0, i;
	
	for (i = 0; string[i]; i++)
		value = value * 10 + string[i] - 0x30;
	
	*decimal = value;
}


// Arbitrary delay function
void delay (int max) {
	int i;
	for (i = 0; i < max; i++) {
		printf("Waiting %d", i);
	}
}
