`timescale 1ns/1ps

/*

File: Arbiter.v

Purpose:
- Instantiates FIFO for each gfx operation to store pixel data (x, y, RGB)
- Performs round-robin arbitration among FIFO's to write pixel data to memory

*/

module Arbiter(
	clk,
	reset,
	soft_reset,

	enq1,
	data_in1,
	full1,

	enq2,
	data_in2,
	full2,

	enq3,
	data_in3,
	full3,

	enq4,
	data_in4,
	full4,

	rtr,
	data_out,
	valid_out
);

	parameter width = 44;	//width of data

	input clk;
	input reset;			//Active H, synchronous reset
	input soft_reset;		//Active H, synchronous Soft reset
	
	//Enqueue of FIFOx
	input enq1;	
	input enq2;	
	input enq3;	
	input enq4;	
	
	//Data in from Operation to FIFOx
	input [width-1:0] data_in1;	
	input [width-1:0] data_in2;	
	input [width-1:0] data_in3;	
	input [width-1:0] data_in4;	

	//Active H, FIFOx is full
	output full1;
	output full2;
	output full3;
	output full4;
	
	input rtr;						//Display Controller is ready to read data
	output [width-1:0] data_out;	//output data from arbiter to mem
	output valid_out;				//active H, data out is valid
	 
	wire nempty1,nempty2,nempty3,nempty4;				//Active H, FIFOx is NOT EMPTY 
	wire empty1,empty2,empty3,empty4;					//Active H, FIFOx is EMPTY 
	wire [width-1:0] data_f1,data_f2,data_f3,data_f4;	//data out from FIFOx 
	reg deq1,deq2,deq3,deq4;							//Active H, dequeue of FIFOx
	
	reg [width-1:0] data_out;
	reg valid_out;
	
	//SM variables
	reg [1:0]	ns,ps;	//next state,present state of control
	parameter	F1=0,	//State FIFO1, giving FIFO2 highest priority
				F2=1,	//State FIFO2, giving FIFO3 highest priority
				F3=2,	//State FIFO3, giving FIFO4 highest priority
				F4=3;	//state FIFO4, giving FIFO1 highest priority
	 
	//control path
	assign nempty1 = !empty1;
	assign nempty2 = !empty2;
	assign nempty3 = !empty3;
	assign nempty4 = !empty4;

	always @(ps or nempty1 or nempty2 or nempty3 or nempty4 or rtr)
	begin
		deq1=0;deq2=0;deq3=0;deq4=0;	//defaults
		ns = ps;
		case(ps)
			F1:	begin
					if(rtr && nempty2) begin
						ns=F2;
						deq2=1;
					end
					else if (rtr && nempty3) begin
						ns=F3;
						deq3=1;
					end
					else if (rtr && nempty4) begin
						ns=F4;
						deq4=1;
					end
					else if (rtr && nempty1) begin
						deq1=1;ns=F1;
					end
					else ns=F1;
				end
			F2:	begin
					if (rtr && nempty3) begin
						ns=F3;
						deq3=1;
					end
					else if(rtr && nempty4) begin
						ns=F4;
						deq4=1;
					end 
					else if(rtr && nempty1) begin
						ns=F1;
						deq1=1;
					end 
					else if (rtr && nempty2) begin 
						deq2=1;
						ns=F2;
					end
					else ns=F2;
				end
			F3:	begin
					if(rtr && nempty4) begin
						ns=F4;
						deq4=1;
					end
					else if(rtr && nempty1) begin
						ns=F1;
						deq1=1;
					end
					else if(rtr && nempty2) begin
						ns=F2;
						deq2=1;
					end
					else if(rtr && nempty3) begin
						ns=F3;
						deq3=1;
					end
					else ns=F3;
				end
			F4:	begin
					if(rtr && nempty1) begin
						ns=F1;
						deq1=1;
					end
					else if(rtr && nempty2) begin
						ns=F2;
						deq2=1;
					end
					else if(rtr && nempty3) begin
						ns=F3;
						deq3=1;
					end
					else if(rtr && nempty4) begin
						ns=F4;
						deq4=1;
					end
					else ns=F4;
				end
			default: ns=2'bxx;
		endcase
	end
	 	
	always @ (posedge clk)
		if(reset || soft_reset)	ps<=F1;
		else					ps<=ns;
	
	//DATA path: set output data/valid signals
	always @(data_f1,data_f2,data_f3,data_f4,deq1,deq2,deq3,deq4) begin
		if (deq1)		begin valid_out=1;data_out=data_f1;end
		else if (deq2)	begin valid_out=1;data_out=data_f2;end
		else if (deq3)	begin valid_out=1;data_out=data_f3;end
		else if (deq4)	begin valid_out=1;data_out=data_f4;end
		else			begin valid_out=0;data_out=0;end
	end
	
	// FIFO's to store pixel data for each operation
	
	Fifo_2 #(width, 16, 4) fifo1 (
		.iCLK(clk),
		.iRST_CLK(reset || soft_reset),
		.iENQUEUE(enq1),
		.iDATAIN(data_in1),
		.oFULL(full1),
		.iDEQUEUE(deq1),
		.oDATAOUT(data_f1),
		.oEMPTY(empty1)
	);

	Fifo_2 #(width, 16, 4) fifo2 (
		.iCLK(clk),
		.iRST_CLK(reset || soft_reset),
		.iENQUEUE(enq2),
		.iDATAIN(data_in2),
		.oFULL(full2),
		.iDEQUEUE(deq2),
		.oDATAOUT(data_f2),
		.oEMPTY(empty2)
	);

	Fifo_2 #(width, 16, 4) fifo3 (
		.iCLK(clk),
		.iRST_CLK(reset || soft_reset),
		.iENQUEUE(enq3),
		.iDATAIN(data_in3),
		.oFULL(full3),
		.iDEQUEUE(deq3),
		.oDATAOUT(data_f3),
		.oEMPTY(empty3)
	);

	Fifo_2 #(width, 16, 4) fifo4 (
		.iCLK(clk),
		.iRST_CLK(reset || soft_reset),
		.iENQUEUE(enq4),
		.iDATAIN(data_in4),
		.oFULL(full4),
		.iDEQUEUE(deq4),
		.oDATAOUT(data_f4),
		.oEMPTY(empty4)
	);
	
endmodule
