`timescale 1ns/1ps

/*

File: Fifo_2.v

Purpose: Final FIFO design used in project.

Author: Hafiz Noordin

*/

module Fifo_2 (
	iCLK,
	iRST_CLK,
	iENQUEUE,
	iDATAIN,
	oFULL,
	iDEQUEUE,
	oDATAOUT,
	oEMPTY
);

parameter			WIDTH = 44;		// x,y,RGB -> 10+10+24 bits
parameter			DEPTH = 16;		// default = 16 deep
parameter			ADDR_WIDTH = 4; // log2(DEPTH)

input				iCLK;
input				iRST_CLK;

// Producer interface
input				iENQUEUE;
input [WIDTH-1:0]	iDATAIN;
output				oFULL;

// Consumer interface
input				iDEQUEUE;
output [WIDTH-1:0]	oDATAOUT;
output				oEMPTY;

wire				oFULL;
wire [WIDTH-1:0]	oDATAOUT;
wire				oEMPTY;

// Read/write pointers to Fifo memory
// - extend to ADDR_WIDTH+1 bits so that MSB of rptr/wptr determines full/empty
reg  [ADDR_WIDTH:0]	rptr;
reg  [ADDR_WIDTH:0]	wptr;
wire [ADDR_WIDTH:0]	rptr_plus1;
wire [ADDR_WIDTH:0]	wptr_plus1;
wire [ADDR_WIDTH:0]	rptr_next;
wire [ADDR_WIDTH:0]	wptr_next;

wire				enqueue_valid;
wire				dequeue_valid;

// Flop Array - FIFO memory
reg [WIDTH-1:0]		flopArray [0:DEPTH-1];
wire [WIDTH-1:0]	flopArray_next;

// debug wires
wire [WIDTH-1:0]	flopArray0 = flopArray[0];
wire [WIDTH-1:0]	flopArray1 = flopArray[1];
wire [WIDTH-1:0]	flopArray2 = flopArray[2];
wire [WIDTH-1:0]	flopArray3 = flopArray[3];
wire [WIDTH-1:0]	flopArray4 = flopArray[4];
wire [WIDTH-1:0]	flopArray5 = flopArray[5];
wire [WIDTH-1:0]	flopArray6 = flopArray[6];
wire [WIDTH-1:0]	flopArray7 = flopArray[7];

// Next pointer value
assign rptr_plus1 = rptr + 1;
assign wptr_plus1 = wptr + 1;

// Valid enq/deq
assign enqueue_valid = iENQUEUE && !oFULL;
assign dequeue_valid = iDEQUEUE && !oEMPTY;

// Next rptr/wptr logic
assign rptr_next = dequeue_valid ? rptr_plus1 : rptr;
assign wptr_next = enqueue_valid ? wptr_plus1 : wptr;

// Mux for flop array input
assign flopArray_next = enqueue_valid ? iDATAIN : flopArray[wptr[ADDR_WIDTH-1:0]];

// Output data
assign oDATAOUT = flopArray[rptr[ADDR_WIDTH-1:0]];

// Logic to determine empty/full based on rptr/wptr
assign oEMPTY = (rptr == wptr);
assign oFULL = (rptr[ADDR_WIDTH-1:0] == wptr[ADDR_WIDTH-1:0]) && (rptr[ADDR_WIDTH] != wptr[ADDR_WIDTH]);

// Flops
always @(posedge iCLK) begin
	// Reset
	if (iRST_CLK) begin
		rptr <= 0;
		wptr <= 0;
	end
	else begin
		rptr <= #0.01 rptr_next;
		wptr <= #0.01 wptr_next;
		flopArray[wptr[ADDR_WIDTH-1:0]] <= #0.01 flopArray_next;
	end
end

// Debug messages (simulation only)
always @(posedge iCLK) begin
	if (iENQUEUE && oFULL)
		$display("FIFO (%t): FIFO FULL", $realtime);
	if (iDEQUEUE && oEMPTY)
		$display("FIFO (%t): FIFO EMPTY", $realtime);
end

endmodule
