/*

File: Line.v

Purpose: Hardware implementation of line drawing algorithm

Author: Hafiz Noordin

Algorithm:
 function line(x0, x1, y0, y1)
     boolean steep := abs(y1 - y0) > abs(x1 - x0)
     if steep then
         swap(x0, y0)
         swap(x1, y1)
     if x0 > x1 then
         swap(x0, x1)
         swap(y0, y1)
     int deltax := x1 - x0
     int deltay := abs(y1 - y0)
     int error := 0
     int ystep
     int y := y0
     if y0 < y1 then ystep := 1 else ystep := -1
     for x from x0 to x1
         if steep then plot(y,x) else plot(x,y)
         error := error + deltay
         if 2error = deltax then
             y := y + ystep
             error := error - deltax
*/

`timescale 1ns/1ps

module Line (
	iCLK,
	iRST,
	iSOFT_RST,

	// Decoder interface
	iREQ,
	iDATA,
	oRTR,

	// output FIFO interface
	iFIFO_FULL,
	oFIFO_WE,
	oFIFO_WD,

	// debug bus
	oDEBUG_VALID,
	oDEBUG_BUS
);

// States for FSM when algorithm active
parameter	S0 = 0;
parameter	S1 = 1;
parameter	S2 = 2;


input			iCLK;
input			iRST;
input			iSOFT_RST;

// Decoder interface
input			iREQ;
input [159:0]	iDATA;
output			oRTR;			// ready to receive (!busy)

// output FIFO interface
input			iFIFO_FULL;
output			oFIFO_WE;
output [43:0]	oFIFO_WD;

// debug bus
output			oDEBUG_VALID;
output [31:0]	oDEBUG_BUS;

reg				oRTR;
reg				oFIFO_WE;
reg [43:0]		oFIFO_WD;

reg				oDEBUG_VALID;
reg [31:0]		oDEBUG_BUS;

wire			busy = !oRTR;

// ----------------------------------------------
// Pipe stage 0: Get input data 
// ----------------------------------------------

reg [9:0]		x0, x1, y0, y1;
reg [23:0]		RGB;
reg				valid_req;

reg [1:0]		state, state_next;

always @(posedge iCLK) begin
	if (iRST || iSOFT_RST) begin
		oRTR <= 1;
		valid_req <= 0;
	end

	// RTS & RTR == 1 -> latch input data
	else if (iREQ && oRTR) begin
		oRTR <= 0;
		RGB <= iDATA[23:0];
		x0 <= iDATA[41:32];
		y0 <= iDATA[73:64];
		x1 <= iDATA[105:96];
		y1 <= iDATA[137:128];
		valid_req <= 1;
	end
	
	// wait for FSM at bottom of pipe to finish
	else if (!oRTR) begin
		valid_req <= 0;
		if (state == S2)
			oRTR <= 1;
	end
end

// ----------------------------------------------
// Pipe stage 1: Calculate abs(x0-x1), abs(y0-y1)
// ----------------------------------------------

wire 			x0_gt_x1 = x0 > x1;
wire 			y0_gt_y1 = y0 > y1;
wire [9:0]		abs_deltax = x0_gt_x1 ? x0 - x1 : x1 - x0;
wire [9:0]		abs_deltay = y0_gt_y1 ? y0 - y1 : y1 - y0;

reg [9:0]		d1_x0, d1_x1, d1_y0, d1_y1;
reg [23:0]		d1_RGB;
reg 			d1_valid_req;

reg 			d1_x0_gt_x1;
reg 			d1_y0_gt_y1;
reg [9:0]		d1_abs_deltax;
reg [9:0]		d1_abs_deltay;

always @(posedge iCLK) begin
	if (iRST || iSOFT_RST) begin
		d1_valid_req <= 0;
		d1_x0 <= 0;
		d1_y0 <= 0;
		d1_x1 <= 0;
		d1_y1 <= 0;
		d1_RGB <= 0;
		d1_x0_gt_x1 <= 0;
		d1_y0_gt_y1 <= 0;
		d1_abs_deltax <= 0;
		d1_abs_deltay <= 0;
	end
	else begin
		// Pass data to next stage
		d1_valid_req <= valid_req;
		d1_x0 <= x0;
		d1_y0 <= y0;
		d1_x1 <= x1;
		d1_y1 <= y1;
		d1_RGB <= RGB;
		d1_x0_gt_x1 <= x0_gt_x1;
		d1_y0_gt_y1 <= y0_gt_y1;
		d1_abs_deltax <= abs_deltax;
		d1_abs_deltay <= abs_deltay;
	end
end

// ----------------------------------------------
// Pipe stage 2: Rearrange x0, x1, y0, y1
// ----------------------------------------------

reg [9:0]		d2_x0, d2_x1, d2_y0, d2_y1;
reg [9:0]		d2_x0_next, d2_x1_next, d2_y0_next, d2_y1_next;
reg [23:0]		d2_RGB;
reg 			d2_valid_req;

reg [9:0]		d2_deltax;
reg [9:0]		d2_abs_deltay;

reg				add_subn;
reg				d1_steep;

wire 			steep = d1_abs_deltay > d1_abs_deltax;
wire			d2_y1_gt_y0 = d2_y1_next > d2_y0_next;
wire [9:0]		d2_deltax_next = d2_x1_next - d2_x0_next;
wire [9:0]		d2_abs_deltay_next = d2_y1_gt_y0 ? d2_y1_next - d2_y0_next : d2_y0_next - d2_y1_next;

// Comb'l logic to rearrange coordinates
// - based on if(steep) and if (x0 > x1) as in algorithm above
// - note thate if steep == 1, we need to check y0 > y1 instead
always @(steep or d1_x0_gt_x1 or d1_y0_gt_y1 or d1_x0 or d1_x1 or d1_y0 or d1_y1) begin
	casex ({steep,d1_x0_gt_x1,d1_y0_gt_y1})
		3'b00x:		{d2_x0_next, d2_x1_next, d2_y0_next, d2_y1_next} =
						{d1_x0, d1_x1, d1_y0, d1_y1};
		3'b01x:		{d2_x0_next, d2_x1_next, d2_y0_next, d2_y1_next} =
						{d1_x1, d1_x0, d1_y1, d1_y0};
		3'b1x0:		{d2_x0_next, d2_x1_next, d2_y0_next, d2_y1_next} =
						{d1_y0, d1_y1, d1_x0, d1_x1};
		3'b1x1:		{d2_x0_next, d2_x1_next, d2_y0_next, d2_y1_next} =
						{d1_y1, d1_y0, d1_x1, d1_x0};
		default: 	{d2_x0_next, d2_x1_next, d2_y0_next, d2_y1_next} = 'bx;

	endcase
end

always @(posedge iCLK) begin
	if (iRST || iSOFT_RST) begin
		d2_valid_req <= 0;
		d2_x0 <= 0;
		d2_y0 <= 0;
		d2_x1 <= 0;
		d2_y1 <= 0;
		d2_RGB <= 0;
		d2_deltax <= 0;
		d2_abs_deltay <= 0;
		add_subn <= 0;
		d1_steep <= 0;
	end
	else begin
		// Pass data to next stage
		d2_valid_req <= d1_valid_req;
		d2_x0 <= d2_x0_next;
		d2_y0 <= d2_y0_next;
		d2_x1 <= d2_x1_next;
		d2_y1 <= d2_y1_next;
		d2_RGB <= d1_RGB;
		d2_deltax <= d2_deltax_next;
		d2_abs_deltay <= d2_abs_deltay_next;
		add_subn <= d2_y1_gt_y0;
		d1_steep <= steep;
	end
end

// ----------------------------------------------
// Pipe stage 3: main loop
// ----------------------------------------------

reg [9:0]	error, error_next;							// error variable
reg [9:0]	x, x_next, y, y_next;						// x,y counters

wire		error_neg = error[9];						// look at sign bit to determine if error < 0
wire [9:0] 	dy_minus_dx = d2_abs_deltay - d2_deltax;
wire		incr_y = !error_neg && ({error + d2_abs_deltay, 1'b0} >= {1'b0, d2_deltax});	// if (2*error) > deltax

reg			fifo_we_next;		// fifo write enable/data
reg [43:0]	fifo_wd_next;

// Comb'l logic to determine next: state, error value, x/y coordinates, fifo write enable/data
always @(state or d2_valid_req or error or x or y or oFIFO_WE or oFIFO_WD or iFIFO_FULL or d2_x0 or d2_x1 or d2_y0 or d1_steep or d2_RGB or d2_deltax or d2_abs_deltay or incr_y or add_subn) begin

	// Defaults
	state_next = state;
	error_next = error;
	x_next = x;
	y_next = y;
	fifo_we_next = oFIFO_WE;
	fifo_wd_next = oFIFO_WD;

	case (state)
		S0:			begin
						// Wait for valid request
						if (d2_valid_req) begin
							state_next = S1;
							x_next = d2_x0;
							y_next = d2_y0;
						end
					end

		S1:			begin
						// Algorithm active state

						// only write pixel data if output fifo ready
						if (!iFIFO_FULL) begin
							fifo_we_next = 1;
							fifo_wd_next = d1_steep ? {y, x, d2_RGB} : {x, y, d2_RGB};

							// Determine next error value
							if (incr_y) begin
								error_next = error + dy_minus_dx;
								y_next = add_subn ? y + 1 : y - 1;
							end
							else begin
								error_next = error + d2_abs_deltay;
							end

							// Done when x reaches final x1
							if (x == d2_x1) begin
								state_next = S2;
							end

							// Increment x counter
							x_next = x + 1;
						end

						// else, stall (no write to fifo)
						else begin
							fifo_we_next = 0;
						end
					end

		S2:			begin
						// state to signal done (oRTR -> 1), reset counters
						state_next = S0;
						error_next = 0;
						x_next = 0;
						y_next = 0;
						fifo_we_next = 0;
						fifo_wd_next = 0;
					end

		default:	begin // unexpected state
						state_next = 'bx;
						error_next = 0;
						x_next = 0;
						y_next = 0;
						fifo_we_next = 0;
						fifo_wd_next = 0;
					end
	endcase
end

// Flops
always @(posedge iCLK) begin
	if (iRST || iSOFT_RST) begin
		state <= S0;
		error <= 0;
		x <= 0;
		y <= 0;
		oFIFO_WE <= 0;
		oFIFO_WD <= 0;
		oDEBUG_VALID <= 0;
		oDEBUG_BUS <= 0;
	end
	else begin
		state <= state_next;
		error <= error_next;
		x <= x_next;
		y <= y_next;
		oFIFO_WE <= fifo_we_next;
		oFIFO_WD <= fifo_wd_next;

		// Debug bus registers
		if (state == S0 && state_next == S1) begin
			oDEBUG_VALID <= 1;
			oDEBUG_BUS <= {2'b0, d2_abs_deltay, 2'b0, d2_deltax, 2'b0, add_subn, d1_steep, state_next, state};
		end
		else if (state == S1) begin
			oDEBUG_VALID <= 1;
			oDEBUG_BUS <= {2'b0, y_next, 2'b0, error_next, 2'b0, add_subn, d1_steep, state_next, state};
		end
		else begin
			oDEBUG_VALID <= 0;
		end
	end
end


endmodule
