`timescale 1ns/1ps
/*

File: OpbInterface.v

Purpose: OPB slave interface for register reads/writes.  Stores registers to accept request operations and data.

Adapted from: OpbInterface.v from module m05

Author: Hafiz Noordin

*/

module OpbInterface(
	OPB_ABus,
	OPB_BE,
	OPB_Clk,
	OPB_DBus,
	OPB_RNW,
	OPB_Rst,
	OPB_select,
	OPB_seqAddr,
	OPB_hwXfer, 
	OPB_fwXfer, 
	Sln_xferAck, 
	Sln_hwAck, 
	Sln_fwAck, 
	Sln_errAck, 
	Sln_toutSup, 
	Sln_retry, 
	Sln_DBusEn, 
	Sln_DBus,
	status,
	input_data,
	blit_rtr,
	line_rtr,
	char_rtr,
	gfx2d_rtr,
	arb_line_full,
	arb_blit_full,
	arb_char_full,
	arb_pixel_full,
	arb_mem_rtr,
	debug_fifo_data,
	debug_fifo_empty
);
   
// Address parameters
parameter C_BASEADDR = 32'h80040000;
parameter C_HIGHADDR = 32'h800400ff; // 256 byte space
parameter C_OPB_AWIDTH = 32;
parameter C_OPB_DWIDTH = 32;

//status Byte Constants
//`define SOFT_RESET_BIT 0
//`define OP_BIT 2:1
`define OP_BIT 2:0
`define REQ_BIT 3

// OPB interface
input						OPB_Clk;			// clock
input						OPB_Rst;			// hard reset
input [C_OPB_AWIDTH-1:0]	OPB_ABus;			// address bus
input [C_OPB_DWIDTH-1:0]	OPB_DBus;			// data bus
input [C_OPB_DWIDTH/8-1:0]	OPB_BE;				// byte enable for data bus
input						OPB_RNW;			// read=1, write=0
input						OPB_select;
input						OPB_seqAddr;
input 						OPB_hwXfer;
input						OPB_fwXfer;
output						Sln_xferAck;
output						Sln_hwAck;
output						Sln_fwAck;
output						Sln_errAck;
output						Sln_toutSup;
output						Sln_retry;
output						Sln_DBusEn;
output [C_OPB_DWIDTH-1:0]	Sln_DBus;			// Assert data when register read requested, else assert 0's
reg [C_OPB_DWIDTH-1:0]		Sln_DBus;

// Status register - stores input request and op code
output [C_OPB_DWIDTH-1:0]	status; 
reg [C_OPB_DWIDTH-1:0]		status; 

// Data Register - stores input data
output [159:0] 	    		input_data;
reg [159:0] 	    		input_data;

// Status bits from 2d blocks
input						blit_rtr;
input						line_rtr;
input						char_rtr;
input						gfx2d_rtr;

// Status bits from arbiter
input						arb_line_full;
input						arb_blit_full;
input						arb_char_full;
input						arb_pixel_full;
input						arb_mem_rtr;

// Debug data
input [31:0]				debug_fifo_data;
input						debug_fifo_empty;

//state: 0 -> Waiting for Valid Address, 1-> Doing Read, 2->Doing Write, 3-> Error
reg [1:0]					state;

wire						ValidAddr;			// OPB_Abus is in our range
wire [C_OPB_DWIDTH-1:0]		OPB_DBus_BE;		// Masked OPB data

`ifdef SIM
initial begin
	state <= 0;
	status <= 0;
	input_data <= 0;
end
`endif

assign ValidAddr = (OPB_ABus <= C_HIGHADDR) & (OPB_ABus >= C_BASEADDR);

assign Sln_hwAck = 0;
assign Sln_toutSup = 0;

// Assert data on output bus when doing register read
assign Sln_DBusEn = state === 1;

// Acknowledge read/write request
assign Sln_xferAck = state === 1 | state === 2;
assign Sln_fwAck = state === 1 | state === 2;

// Output error signals if problem
assign Sln_retry = state === 3;
assign Sln_errAck = state === 3;

// Mask OPB data
assign OPB_DBus_BE 		= {	(OPB_BE[3] ? OPB_DBus[31:24] : 8'b0),
							(OPB_BE[2] ? OPB_DBus[23:16] : 8'b0),
							(OPB_BE[1] ? OPB_DBus[15: 8] : 8'b0),
							(OPB_BE[0] ? OPB_DBus[ 7: 0] : 8'b0) };

always @ (posedge OPB_Clk) begin

	if (OPB_Rst) begin
		state <= 0;
		status <= 0;
		input_data <= 0;
	end

	else begin
		if(state === 0) begin
			// State 0: idle
			if(OPB_select & ValidAddr) begin
				if(OPB_hwXfer & OPB_fwXfer) begin

					// Register read request
					if(OPB_RNW) begin
						state <= 1;

						// Assert status signals and current register data onto bus
						case (OPB_ABus)
							C_BASEADDR      :	Sln_DBus <= {19'b0, arb_pixel_full, arb_char_full, arb_line_full, arb_blit_full, arb_mem_rtr, gfx2d_rtr, char_rtr, line_rtr, blit_rtr, status[3:0]};
							C_BASEADDR + 4  :	Sln_DBus <= input_data[31:0];		//RGB
							C_BASEADDR + 8  :	Sln_DBus <= input_data[63:32];		//x0
							C_BASEADDR + 12 :	Sln_DBus <= input_data[95:64];		//y0
							C_BASEADDR + 16 :	Sln_DBus <= input_data[127:96];		//x1
							C_BASEADDR + 20 :	Sln_DBus <= input_data[159:128];	//y1
							C_BASEADDR + 24 :	Sln_DBus <= debug_fifo_empty ? 32'hdeadbeef : debug_fifo_data;
							default:			Sln_DBus <= 0;
						endcase
					end

					// Register write request
					else begin
						state <= 2;
						Sln_DBus <= 0;
						case (OPB_ABus)
							C_BASEADDR      :	status <= OPB_DBus_BE;
							C_BASEADDR + 4  :	input_data[31:0]   <= OPB_DBus_BE;
							C_BASEADDR + 8  :	input_data[63:32]  <= OPB_DBus_BE;
							C_BASEADDR + 12 :	input_data[95:64]  <= OPB_DBus_BE;
							C_BASEADDR + 16 :	input_data[127:96] <= OPB_DBus_BE;
							C_BASEADDR + 20 :	input_data[159:128] <= OPB_DBus_BE;
						endcase
					end 
				end

				// Error in request
				else begin
					state <= 3; 
					Sln_DBus <= 0;
				end
			end
		end

		else if (state === 1 | state === 2 | state === 3) begin
			// State 1: prev cycle was read
			// State 2: prev cycle was write
			// State 3: prev cycle was error
		
			// Go back to idle state
			state <= 0; 

			//Keep these bits high for only 1 cycle.
			//status[`SOFT_RESET_BIT] <= 0;
			status[`REQ_BIT] <= 0;

			Sln_DBus <= 0;
		end
	end
end

endmodule
