/*

File: gfx2d.v

Purpose: top level file for 2D Graphics Engine

Author: Hafiz Noordin

*/

`timescale 1ns/1ps

module gfx2d (
	OPB_ABus,
	OPB_BE,
	OPB_Clk,
	OPB_DBus,
	OPB_RNW,
	OPB_Rst,
	OPB_select,
	OPB_seqAddr,
	OPB_hwXfer, 
	OPB_fwXfer, 
	Sln_xferAck, 
	Sln_hwAck, 
	Sln_fwAck, 
	Sln_errAck, 
	Sln_toutSup, 
	Sln_retry, 
	Sln_DBusEn, 
	Sln_DBus,

	pixel_clock,
	reset,
	user_vga_addr,
	user_vga_read,
	user_vga_write,
	user_vga_write_byte_a,
	user_vga_write_byte_b,
	user_vga_write_byte_c,
	user_vga_write_byte_d,
	user_vga_enable,
	user_vga_read_data,
	user_vga_write_data,
	user_access_ok,
	user_clock_enable,
	user_load_vga_addr
);

parameter C_BASEADDR = 32'h80040000;
parameter C_HIGHADDR = 32'h800400ff; // 256 byte space
parameter C_OPB_AWIDTH = 32;
parameter C_OPB_DWIDTH = 32;

// Screen resolution
`define RESOLUTION_H 640
`define RESOLUTION_V 480

// Status Byte Constants
`define OP_BIT 2:0
`define REQ_BIT 3

// 2D op codes
parameter OP_LINE = 0;
parameter OP_BLIT = 1;
parameter OP_CHAR = 2;
parameter OP_PIXEL= 3;
parameter OP_DBG  = 4;

parameter data_width = 160;		// width of data from OPB bus
parameter op_width = 3;			// width of op from OPB bus

// OPB Bus interface
input						OPB_Clk;
input						OPB_Rst;
input [C_OPB_AWIDTH-1:0]	OPB_ABus;
input [C_OPB_DWIDTH-1:0]	OPB_DBus;
input [C_OPB_DWIDTH/8-1:0]	OPB_BE;
input						OPB_RNW;
input						OPB_select;
input						OPB_seqAddr;
input 						OPB_hwXfer;
input						OPB_fwXfer;
output						Sln_xferAck;
output						Sln_hwAck;
output						Sln_fwAck;
output						Sln_errAck;
output						Sln_toutSup;
output						Sln_retry;
output						Sln_DBusEn;
output [C_OPB_DWIDTH-1:0]	Sln_DBus;

// Display interface
input						pixel_clock		/* synthesis syn_noclockbuf = 1 */; // pixel clock input
input						reset;					// hard reset
output [18:0]				user_vga_addr;
output						user_vga_read;
output						user_vga_write;
output						user_vga_write_byte_a;
output						user_vga_write_byte_b;
output						user_vga_write_byte_c;
output						user_vga_write_byte_d;
output						user_vga_enable;
input  [31:0]				user_vga_read_data;
output [31:0]				user_vga_write_data;
input						user_access_ok;
output						user_clock_enable;
output						user_load_vga_addr;		// flag to load the address specified by the user_vga_add bus 

// OPB interface <-> Decoder 
wire [C_OPB_DWIDTH-1:0]		status; 					// status register from OPB interface
wire						input_req;					// request bit from status register
wire [op_width-1:0]			input_op;					// op code from status register
wire						soft_reset;					// soft_reset bit from status register (CURRENTLY UNUSED)
wire [data_width-1:0]		input_data;					// data bus


// Decoder <-> 2D modules 
wire [data_width-1:0]		decoder_data;								// data bus
wire						decoder_full;								// full signal from decoder input fifo
wire						line_req, blit_req, char_req, pixel_req;	// Request signals from decoder
wire						line_rtr, blit_rtr, char_rtr, pixel_rtr;	// RTR (ready to receive) signals from each 2D module

// 2D modules <-> Arbiter <-> Memory
wire						line_full, blit_full, char_full, pixel_full;	// Output FIFO full signals for each 2D module
wire						we, line_we, blit_we, char_we, pixel_we;		// Write enable
wire [43:0]					wd, line_wd, blit_wd, char_wd, pixel_wd;		// Write data
wire						mem_rtr;										// Memory ready to receive pixel data

// Debug signals
wire						line_debug_valid;
wire						debug_fifo_full;
wire [31:0]					line_debug_data;

wire						debug_fifo_deq;
wire [31:0]					debug_fifo_data;
wire						debug_fifo_empty;

// Data path

//assign soft_reset = status[`SOFT_RESET_BIT]; // previously had a register bit for soft reset, now just tie low and use on-board (hard) reset
assign soft_reset = 0;


// Input signals from user - request, op (via register writes)
assign input_req = status[`REQ_BIT];
assign input_op = status[`OP_BIT];

// Debug operation - dequeue data from debug fifo
assign debug_fifo_deq = input_req && input_op == OP_DBG;

// Pixel operation: pass input data directly into output fifo
assign pixel_wd = {decoder_data[41:32], decoder_data[73:64], decoder_data[23:0]}; // x, y, RGB
assign pixel_rtr = !pixel_full;

// OPB bus interface: for register reads/writes
OpbInterface uOpbInterface (
	.OPB_ABus(OPB_ABus),
	.OPB_BE(OPB_BE),
	.OPB_Clk(OPB_Clk),
	.OPB_DBus(OPB_DBus),
	.OPB_RNW(OPB_RNW),
	.OPB_Rst(OPB_Rst),
	.OPB_select(OPB_select),
	.OPB_seqAddr(OPB_seqAddr),
	.OPB_hwXfer(OPB_hwXfer) ,
	.OPB_fwXfer(OPB_fwXfer) ,
	.Sln_xferAck(Sln_xferAck) ,
	.Sln_hwAck(Sln_hwAck) ,
	.Sln_fwAck(Sln_fwAck) ,
	.Sln_errAck(Sln_errAck) ,
	.Sln_toutSup(Sln_toutSup) ,
	.Sln_retry(Sln_retry) ,
	.Sln_DBusEn(Sln_DBusEn) ,
	.Sln_DBus(Sln_DBus),
	.status(status),
	.input_data(input_data),
	.blit_rtr(blit_rtr),
	.line_rtr(line_rtr),
	.char_rtr(char_rtr),
	.gfx2d_rtr(!decoder_full),
	.arb_line_full (line_full),
	.arb_blit_full (blit_full),
	.arb_char_full (char_full),
	.arb_pixel_full(pixel_full),
	.arb_mem_rtr (user_access_ok),
	.debug_fifo_data (debug_fifo_data),
	.debug_fifo_empty (debug_fifo_empty)
);

// Decoder: parses input request and passes data to respective 2D module
Decoder uDecoder (
	// Inputs
	.clk(OPB_Clk),
	.reset(OPB_Rst),
	.soft_reset(soft_reset),
	.data(input_data),
	.op(input_op),
	.enq(input_req),
	.busy_line(!line_rtr),
	.busy_blit(!blit_rtr),
	.busy_char(!char_rtr),
	.busy_pixel(!pixel_rtr),

	// Outputs
	.full(decoder_full),
	.data_out(decoder_data),
	.valid_line(line_req),
	.valid_blit(blit_req),
	.valid_char(char_req),
	.valid_pixel(pixel_req)
);

// Line operation
Line uLine(
	.iCLK(OPB_Clk),
	.iRST(OPB_Rst),
	.iSOFT_RST(soft_reset),

	// Decoder interface
	.iREQ(line_req),
	.iDATA(decoder_data),
	.oRTR(line_rtr),

	// output FIFO interface
	.iFIFO_FULL (line_full),
	.oFIFO_WE (line_we),
	.oFIFO_WD (line_wd),

	.oDEBUG_VALID (line_debug_valid),
	.oDEBUG_BUS (line_debug_data)
);

// Blit operation
Blit uBlit(
	.iCLK(OPB_Clk),
	.iRST(OPB_Rst),
	.iSOFT_RST(soft_reset),

	// Decoder interface
	.iREQ(blit_req),
	.iDATA(decoder_data),
	.oRTR(blit_rtr),

	// output FIFO interface
	.iFIFO_FULL (blit_full),
	.oFIFO_WE (blit_we),
	.oFIFO_WD (blit_wd)
);

// Character draw operation
Char uChar (
	.clk(OPB_Clk),
	.reset(OPB_Rst),
	.soft_reset(soft_reset),

	.req(char_req),
	.data(decoder_data),
	.rtr(char_rtr),

	.fifo_full(char_full),
	.fifo_we(char_we),
	.fifo_wd(char_wd)
);

// Arbiter: organizes writes to frame buffer
Arbiter uArbiter (
	.clk(OPB_Clk),
	.reset(OPB_Rst),
	.soft_reset(soft_reset),

	.enq1(line_we),
	.data_in1(line_wd),
	.full1(line_full),

	.enq2(blit_we),
	.data_in2(blit_wd),
	.full2(blit_full),

	.enq3(char_we),
	.data_in3(char_wd),
	.full3(char_full),

	.enq4(pixel_req),
	.data_in4(pixel_wd),
	.full4(pixel_full),

	.rtr (mem_rtr),
	.valid_out(we),
	.data_out(wd)
);

// Memory interface: from Arbiter output to ZBT controller

wire [9:0]					x;
wire [9:0]					y;
wire [23:0]					RGB;

assign {x,y,RGB} = wd;
//assign user_vga_addr = y * `RESOLUTION_H + x;
assign user_vga_addr = {y, 9'b0} + {2'b0, y, 7'b0} + {9'b0, x}; // equivalent to: y * `RESOLUTION_H + x

assign user_vga_read = 0;			// doing write only

assign user_vga_write_data = {8'b0, RGB};	 // 24-bit RGB data (most-sig byte of write bus = 0)
assign user_vga_write_byte_d = 0;	

// assert signals required to do write
assign {user_vga_write, user_vga_write_byte_a, user_vga_write_byte_b, user_vga_write_byte_c, user_vga_enable, user_clock_enable, user_load_vga_addr} = we ? {7{1'b1}} : 0;

assign mem_rtr = user_access_ok;	// based on SVGA timing generation


// Debug fifo - capture line debug bus for debug purposes
Fifo_2 #(32, 128, 7) debug_fifo (
	.iCLK(OPB_Clk),
	.iRST_CLK(OPB_Rst || soft_reset),
	.iENQUEUE(line_debug_valid && !debug_fifo_full),
	.iDATAIN(line_debug_data),
	.oFULL(debug_fifo_full),
	.iDEQUEUE(debug_fifo_deq),
	.oDATAOUT(debug_fifo_data),
	.oEMPTY(debug_fifo_empty)
);

// Memory Interface - experiment to write to framebuffer and read back data to  verify correct data written
/*
MemIF uMemIF (
	.iCLK(OPB_Clk),
	.iRST(OPB_Rst),
	.iSOFT_RST(soft_reset),

	.iVALID(we),
	.iDATA(wd),
	.oRTR(mem_rtr),

	.user_access_ok(user_access_ok),
	.user_vga_read_data(user_vga_read_data),

	.user_clock_enable(user_clock_enable),
	.user_load_vga_addr(user_load_vga_addr),
	.user_vga_addr(user_vga_addr),
	.user_vga_enable(user_vga_enable),
	.user_vga_read(user_vga_read),
	.user_vga_write(user_vga_write),
	.user_vga_write_byte_a(user_vga_write_byte_a),
	.user_vga_write_byte_b(user_vga_write_byte_b),
	.user_vga_write_byte_c(user_vga_write_byte_c),
	.user_vga_write_byte_d(user_vga_write_byte_d),
	.user_vga_write_data(user_vga_write_data)
);
*/

// Test module to display single colour on screen
/*
Display uDisplay (
	.pixel_clock(pixel_clock),
	.reset(reset),
	.user_vga_addr(user_vga_addr),
	.user_vga_read(user_vga_read),
	.user_vga_write(user_vga_write),
	.user_vga_write_byte_a(user_vga_write_byte_a),
	.user_vga_write_byte_b(user_vga_write_byte_b),
	.user_vga_write_byte_c(user_vga_write_byte_c),
	.user_vga_write_byte_d(user_vga_write_byte_d),
	.user_vga_enable(user_vga_enable),
	.user_vga_read_data(user_vga_read_data),
	.user_vga_write_data(user_vga_write_data),
	.user_access_ok(user_access_ok),
	//.user_clock(user_clock),
	//.user_ram_clock(user_ram_clock),
	.user_clock_enable(user_clock_enable),
	.user_load_vga_addr(user_load_vga_addr),
	.iRGB_DATA(input_data[23:0])
);
*/


endmodule
